/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/backend/cpu/operator/GridSampleImpl.hpp"

#include <functional>
#include <vector>

#include "aidge/backend/cpu/data/GetCPUPtr.h"
#include "aidge/backend/cpu/operator/GridSampleImpl_forward_kernels.hpp"
#include "aidge/operator/GridSample.hpp"
#include "aidge/utils/Types.h"

Aidge::Elts_t Aidge::GridSampleImpl_cpu::getNbRequiredProtected(IOIndex_t /*inputIdx*/) const {
    // this implementation can be in-place
    return Elts_t::DataElts(0);
}

void Aidge::GridSampleImpl_cpu::forward() {
    const auto& op_ = static_cast<const GridSample_Op&>(mOp);

    // Find the correct kernel type
    const auto outputDataType = op_.getOutput(0)->dataType();

    const Registrar<GridSampleImpl1DForward_cpu>::registrar_key registrarKey = {
        op_.getInput(0)->dataType(),
        outputDataType};

    std::function<void(const GridSample_Op&,
                            const std::shared_ptr<Tensor>&,
                            const std::shared_ptr<Tensor>&,
                            const std::shared_ptr<Tensor>&)> kernelFunc;

    const std::size_t nbSpatialFeat = op_.getInput(0)->nbDims();
    switch (nbSpatialFeat)
    {
    case 1:
        kernelFunc = Registrar<GridSampleImpl1DForward_cpu>::create(registrarKey);
        break;
    case 2:
        kernelFunc = Registrar<GridSampleImpl2DForward_cpu>::create(registrarKey);
        break;
    default:
        AIDGE_THROW_OR_ABORT(std::runtime_error, "No CPU {} kernel available for {} dimensions.", op_.type(), nbSpatialFeat);
        break;
    }

    // Convert input data (no overhead if not needed!)
    // TODO: right now, if needed, memory will be allocated/deallocated at each
    // call to forward(). We might put the following shared_ptr as members of
    // this class to avoid that.
    std::shared_ptr<Tensor> input0Fallback, input1Fallback;
    const auto& input0 = std::make_shared<Tensor>(op_.getInput(0)->refCastFrom(input0Fallback, *op_.getOutput(0)));
    const auto& input1 = std::make_shared<Tensor>(op_.getInput(1)->refCastFrom(input1Fallback, *op_.getOutput(0)));

    // Call kernel
    kernelFunc(op_,
            input0, // input
            input1, // grid
            op_.getOutput(0) // output
            );
}
